/* RCS: $Id: bineph2fits.c,v 3.0 1998/08/25 19:26:49 arots Exp arots $ */
/*-----------------------------------------------------------------------
 *
 *  bineph2fits.c
 *
 *  Date: 7 February 1997
 *  FITS version
 *
 *  Arnold Rots, USRA
 *
 *  A program to convert binary JPL ephemeris files to FITS.
 *  Has to run on a big-endian machine because of the nature
 *  of JPL's "unix" files.
 *
 *  Usage:
 *    bineph2fits <recordlength> <jdstart> <jdend> <filename1> [<filename2> [...]]
 *       <recordlength>  The length of the ephemeris records in ints;
 *                       1652 for DE-200, 2036 for DE-405, 1456 for DE-406.
 *       <jdstart>       first JD to be included; =0: start of first file.
 *       <jdend>         last JD to be included; =0: end of last file.
 *       <filename1>     input file name; at least one is required;
 *                       any number may be given, as long as they are
 *                       consecutive and in order.
 *
 *  Output: JPLEPH
 *
 *  Output is a bare-bones FITS file consisting of three binary
 *  table extensions DE1, DE2, and DE3.
 *  DE1 contains a table of the ephemeris constants and their names.
 *  The two columns are CNAME (A7) and CVAL (D).
 *  DE2 contains the information on the structure of the coefficient buffer:
 *  There are 13 rows and three (I) columns: IPTR, NCF, NA.
 *  DE3 consists of a single column, with each cell containing a vector
 *  with a full buffer of Chebyshev coefficients.
 *
 *----------------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fitsio.h>
#include <longnam.h>

int main (int argc, char **argv) 
{
  FILE *binfile ;
  fitsfile *FF ;
  char *ephemFile = "JPLEPH" ;
  int i, j, k ;
  char **cnam ;
  double cval[400] ;
  char ttltmp[252], cnamtmp[2400] ;
  double ss[3] ;
  int ipttmp[40] ;
  int recbytlen, recdbllen, recintlen ;
  double au, ss1, ss2, ss3, ss3inv, clight, emrat ;
  int denum, group, nc ;
  int ncoef, nrow ;
  double t1, t2, rt1, rt2 ;
  int iptr[13], ncf[13], na[13] ;
  double *buffer ;
  char line[100] ;
  char *com1 ;
  char *com2 ;
  char com3[100] ;
  char com4[100] ;
  char com5[100] ;
  int error = 0 ;
  
  char *ttype1[2] = {"Cname", "Cvalue"} ;
  char *tform1[2] = {"6A", "D"} ;
  char *tunit1[2] = {"", ""} ;
  char *extde1 = "DE1" ;
  
  char *ttype2[4] = {"Object", "Pointer", "NumCoeff", "NumSubIntv"} ;
  char *tform2[4] = {"22A", "I", "I", "I"} ;
  char *tunit2[4] = {"", "", "", ""} ;
  char *extde2 = "DE2" ;
  char *objects[13] = {"Mercury", "Venus", "Earth-Moon Barycenter", "Mars",
		       "Jupiter", "Saturn", "Uranus", "Neptune", "Pluto",
		       "Moon (geocentric)", "Sun", "Nutations", "Librations"} ;
  
  char *ttype3 = {"ChebCoeffs"} ;
  char tform3[12] ;
  char *ttemp ;
  char *tunit3 = "" ;
  char *extde3 = "DE3" ;

  double tlast ;
  int first ;
  int dowrite ;

  ttemp =tform3 ;

/*
 *    Get inputs
 */
  argc-- ;
  if ( !argc )
    error = 1 ;
  if ( !error ) {
    recintlen = atoi (*(++argv)) ;
    if ( recintlen <= 0 )
      error = -1 ;
    recbytlen = recintlen * 4 ;
    recdbllen = recintlen / 2 ;
    argc-- ;
    if ( !argc )
      error = 2 ;
  }
  if ( !error ) {
    t1 = atof (*(++argv)) ;
    if ( t1 < 0.0 )
      t1 = 0.0 ;
    argc-- ;
    if ( !argc )
      error = 3 ;
  }
  if ( !error ) {
    t2 = atof (*(++argv)) ;
    argc-- ;
    if ( t2 <= t1 )
      t2 = 1.0e20 ;
    if ( !argc )
      error = 4 ;
  }
  if ( error ) {
    fprintf (stderr, "bineph2fits - Error %d; Usage:\n", error) ;
    fprintf (stderr, "   bineph2fits <recordlength> <jdstart> <jdend> <filename1> [<filename2> [...]]\n") ;
    fprintf (stderr, "      <recordlength>  The length of the ephemeris records in ints;\n") ;
    fprintf (stderr, "                      1652 for DE-200, 2036 for DE-405, 1456 for DE-406.\n") ;
    fprintf (stderr, "      <jdstart>       first JD to be included; =0: start of first file.\n") ;
    fprintf (stderr, "      <jdend>         last JD to be included; =0: end of last file.\n") ;
    fprintf (stderr, "      <filename1>     input file name; at least one is required;\n") ;
    fprintf (stderr, "                      any number may be given, as long as they are\n") ;
    fprintf (stderr, "                      consecutive and in order.\n") ;
    exit (error) ;
  }

/*
 *    First, the end of the range
 */
  binfile = fopen (argv[argc], "r") ;
  if ( !binfile ) {
    fprintf (stderr, "bineph2fits - could not open file %s\n", argv[argc-1]) ;
    exit (5) ;
  }

  if ( fread (ttltmp, sizeof(char), 252, binfile) != 252 ) {
    fprintf (stderr, "bineph2fits - could not read TTL\n") ;
    exit (6) ;
  }

  strncpy (com3, ttltmp+168, 84) ;
  com3[68] = 0 ;

  fclose (binfile) ;

/*
 *    Header
 */
  binfile = fopen (*(++argv), "r") ;
  if ( !binfile ) {
    fprintf (stderr, "bineph2fits - could not open file %s\n", *argv) ;
    exit (5) ;
  }

  if ( fread (ttltmp, sizeof(char), 252, binfile) != 252 ) {
    fprintf (stderr, "bineph2fits - could not read TTL\n") ;
    exit (6) ;
  }

  if ( fread (cnamtmp, sizeof(char), 2400, binfile) != 2400 ) {
    fprintf (stderr, "bineph2fits - could not read CNAM\n") ;
    exit (7) ;
  }

  if ( fread (ss, sizeof(double), 3, binfile) != 3 ) {
    fprintf (stderr, "bineph2fits - could not read SS\n") ;
    exit (8) ;
  }

  if ( fread (&nc, sizeof(int), 1, binfile) != 1 ) {
    fprintf (stderr, "bineph2fits - could not read NCON\n") ;
    exit (9) ;
  }

  if ( fread (&au, sizeof(double), 1, binfile) != 1 ) {
    fprintf (stderr, "bineph2fits - could not read AU\n") ;
    exit (10) ;
  }

  if ( fread (&emrat, sizeof(double), 1, binfile) != 1 ) {
    fprintf (stderr, "bineph2fits - could not read EMRAT\n") ;
    exit (11) ;
  }

  if ( fread (ipttmp, sizeof(int), 40, binfile) != 40 ) {
    fprintf (stderr, "bineph2fits - could not read IPT\n") ;
    exit (12) ;
  }

/*
 *    Read constants values
 */
  if ( fseek (binfile, recbytlen, SEEK_SET) ) {
    fprintf (stderr, "bineph2fits - could not get to CVAL\n") ;
    exit (13) ;
  }

  if ( fread (cval, sizeof(double), 400, binfile) != 400 ) {
    fprintf (stderr, "bineph2fits - could not read EMRAT\n") ;
    exit (14) ;
  }

  fclose (binfile) ;

/*
 *    Set the items for the first two extensions
 */
  ncoef = recdbllen ;

  com1 = ttltmp ;
  com2 = ttltmp + 84 ;
  com1[68] = 0 ;
  com2[68] = 0 ;
  if ( t1 > 0.0 )
    sprintf (com4, "         or: JED=  %f (whichever is greater)", t1) ;
  else
    *com4 = 0 ;
  if ( t2 < 1.0e19 )
    sprintf (com5, "         or: JED=  %f (whichever is less)", t2) ;
  else
    *com5 = 0 ;
  
  ss1 = ss[0] ;
  ss2 = ss[1] ;
  ss3 = ss[2] ;

  cnam = (char **) malloc (nc * sizeof(char*)) ;
  *cnam = (char *) malloc (nc * 7 * sizeof(char)) ;
  for (i=0; i<nc; i++) {
    cnam[i] = (*cnam) + i*7 ;
    strncpy (cnam[i], cnamtmp+i*6, 6) ;
    cnam[i][6] = 0 ;
  }

  for (i=0; i<12; i++) {
    iptr[i] = ipttmp[i*3] ;
    ncf[i] = ipttmp[i*3+1] ;
    na[i] = ipttmp[i*3+2] ;
  }
  denum = ipttmp[36] ;
  iptr[12] = ipttmp[37] ;
  ncf[12] = ipttmp[38] ;
  na[12] = ipttmp[39] ;


/*
 *    Create FITS file
 */
  i = 0 ;
  fits_create_file (&FF, ephemFile, &error) ;
  if ( error ) {
    fprintf (stderr, "bineph2fits: Could not create file %s (error %d)\n",
	     ephemFile, error) ;
    fprintf (stderr, "             (does it exist already?") ;
    fprintf (stderr, " do you have write privileges?)\n") ;
    exit (error) ;
  }
  fits_create_img (FF, 8, 0, (long *) &i, &error) ;
  fits_write_date (FF, &error) ;
  fits_write_comment (FF, com1, &error) ;
  fits_write_comment (FF, com2, &error) ;
  if ( *com4 )
    fits_write_comment (FF, com4, &error) ;
  fits_write_comment (FF, com3, &error) ;
  if ( *com5 )
    fits_write_comment (FF, com5, &error) ;
  fits_write_key (FF, TSTRING, "CREATOR", "bineph2fits", "File created by bineph2fits", &error) ;
  fits_write_chksum (FF, &error) ;
  if ( error ) {
    fprintf (stderr, "bineph2fits: Failed writing primary header\n") ;
    fprintf (stderr, "bineph2fits: Error = %d\n", error) ;
    exit (error) ;
  }
  

/*
 *    First extension
 */
  fits_create_tbl (FF, BINARY_TBL, nc, 2, ttype1, tform1, tunit1, extde1, &error) ;
  fits_write_date (FF, &error) ;
  fits_write_comment (FF, com1, &error) ;
  fits_write_comment (FF, com2, &error) ;
  if ( *com4 )
    fits_write_comment (FF, com4, &error) ;
  fits_write_comment (FF, com3, &error) ;
  if ( *com5 )
    fits_write_comment (FF, com5, &error) ;
  fits_write_key (FF, TSTRING, "CREATOR", "bineph2fits", "File created by bineph2fits", &error) ;
  fits_modify_comment (FF, "TTYPE1", "Names of constants", &error) ;
  fits_modify_comment (FF, "TTYPE2", "Values of constants", &error) ;
  fits_modify_comment (FF, "EXTNAME", "Table of JPL Ephemeris constants", &error) ;
  fits_write_col_str (FF, 1, 1, 1, nc, cnam, &error) ;
  fits_write_col (FF, TDOUBLE, 2, 1, 1, nc, cval, &error) ;
  fits_write_chksum (FF, &error) ;
  if ( error ) {
    fprintf (stderr, "bineph2fits: Failed writing first extension\n") ;
    fprintf (stderr, "bineph2fits: Error = %d\n", error) ;
    exit (error) ;
  }
  fprintf (stderr, "bineph2fits: wrote extension 1\n") ;
  

/*
 *    Second extension
 */
  fits_create_tbl (FF, BINARY_TBL, 13, 4, ttype2, tform2, tunit2, extde2, &error) ;
  fits_write_date (FF, &error) ;
  fits_write_comment (FF, "This table describes the structure of the Chebyshev coefficients",
		      &error) ;
  fits_write_comment (FF, "records in the JPL Ephemeris; one row for each solar system object",
		       &error) ;
  fits_write_comment (FF, com1, &error) ;
  fits_write_comment (FF, com2, &error) ;
  if ( *com4 )
    fits_write_comment (FF, com4, &error) ;
  fits_write_comment (FF, com3, &error) ;
  if ( *com5 )
    fits_write_comment (FF, com5, &error) ;
  fits_write_key (FF, TSTRING, "CREATOR", "bineph2fits", "File created by bineph2fits", &error) ;
  fits_modify_comment (FF, "TTYPE1", "Solar system object", &error) ;
  fits_modify_comment (FF, "TTYPE2",
		       "Pointer for object's coefficients in record", &error) ;
  fits_modify_comment (FF, "TTYPE3", 
		       "Number of Chebyshev coefficients for object", &error) ;
  fits_modify_comment (FF, "TTYPE4",
		       "Number of time sub-intervals for object", &error) ;
  fits_modify_comment (FF, "EXTNAME",
		       "Description of Chebyshev coefficients records", &error) ;
  fits_write_col_str (FF, 1, 1, 1, 13, objects, &error) ;
  fits_write_col (FF, TLONG, 2, 1, 1, 13, iptr, &error) ;
  fits_write_col (FF, TLONG, 3, 1, 1, 13, ncf, &error) ;
  fits_write_col (FF, TLONG, 4, 1, 1, 13, na, &error) ;
  fits_write_chksum (FF, &error) ;
  if ( error ) {
    fprintf (stderr, "bineph2fits: Failed writing second extension\n") ;
    fprintf (stderr, "bineph2fits: Error = %d\n", error) ;
    exit (error) ;
  }
  fprintf (stderr, "bineph2fits: wrote extension 2\n") ;
  

/*
 *    Third extension
 */
  sprintf (tform3, "%dD", ncoef) ;
  fits_create_tbl (FF, BINARY_TBL, nc, 1, &ttype3, &ttemp, &tunit3, extde3, &error) ;
  fits_write_comment (FF, "Table of Chebyshev coefficients; the first two elements",
		      &error) ;
  fits_write_comment (FF, " of each record are start and stop time (day) for that record",
		       &error) ;
  fits_modify_comment (FF, "TTYPE1",
		       "Record of Chebyshev coefficients", &error) ;
  fits_modify_comment (FF, "EXTNAME",
		       "Table of Chebyshev coefficients records", &error) ;
  fits_write_date (FF, &error) ;
  fits_write_comment (FF, com1, &error) ;
  fits_write_comment (FF, com2, &error) ;
  if ( *com4 )
    fits_write_comment (FF, com4, &error) ;
  fits_write_comment (FF, com3, &error) ;
  if ( *com5 )
    fits_write_comment (FF, com5, &error) ;
  
  buffer = (double *) malloc ( recdbllen * sizeof(double)) ;

  ss1 = 1.0E20 ;
  ss2 = 0.0 ;
  nrow = 0 ;
  first = 1 ;

/*
 *    Loop over the bin files; do extensive checking
 */
  while ( argc ) {

    binfile = fopen (*argv, "r") ;
    if ( !binfile ) {
      fprintf (stderr, "bineph2fits - could not open file %s\n", *argv) ;
      continue ;
    }

    if ( fseek (binfile, 2652, SEEK_SET) ) {
      fprintf (stderr, "bineph2fits - could not get to JD range in %s\n", *argv) ;
      fclose (binfile) ;
      continue ;
    }

    if ( fread (ss, sizeof(double), 3, binfile) != 3 ) {
      fprintf (stderr, "bineph2fits - could not read SS in %s\n", *argv) ;
      fclose (binfile) ;
      continue ;
    }

    if ( ( ss[0] > t2 ) || ( ss[1] < t1 ) ) {
      fprintf (stderr, "bineph2fits - skipped file %s (out of range)\n", *argv) ;
      fclose (binfile) ;
      continue ;
    }

    if ( first ) {
      first = 0 ;
      ss2 = ss[0] ;
    }

    if ( ss[1] < ss2 ) {
      fprintf (stderr, "bineph2fits - skipped file %s (out of order)\n", *argv) ;
      fclose (binfile) ;
      continue ;
    }

    if ( ss[0] > ss2 ) {
      fprintf (stderr, "bineph2fits - missing rows at %d (file %s)\n",
	       nrow, *argv) ;
    }

    if ( fseek (binfile, 2*recbytlen, SEEK_SET) ) {
      fprintf (stderr, "bineph2fits - could not get to data in %s\n", *argv) ;
      fclose (binfile) ;
      continue ;
    }

/*
 *    Loop over the records
 */
    while ( !error && !(feof(binfile)) ) {
      dowrite = 1 ;
      if ( fread (buffer, sizeof(double), ncoef, binfile) != ncoef )
	continue ;

      if ( buffer[0] < tlast ) {
	dowrite = 0 ;
	fprintf (stderr, "bineph2fits: duplicate row at number %d\n", nrow) ;
      }
      else if ( buffer[0] != tlast ) {
	fprintf (stderr, "bineph2fits: last end: %f, current start: %f\n",
		 tlast, buffer[0]) ;
	tlast = buffer[1] ;
      }
      else
	tlast = buffer[1] ;

      if ( dowrite && ( buffer[0] <= t2 ) && ( buffer[1] >= t1 ) ) {
	nrow++ ;
	fits_write_col (FF, TDOUBLE, 1, nrow, 1, ncoef, buffer, &error) ;
	if ( buffer[0] < ss1 )
	  ss1 = buffer[0] ;
	if ( buffer[1] > ss2 )
	  ss2 = buffer[1] ;
      }
    }

/*
 *    Close the bin file; prepare for the next
 */
    fclose (binfile) ;

    argc-- ;
    argv++ ;
  }

/*
 *    Finish up the FITS file
 */
  fits_write_key (FF, TSTRING, "CREATOR", "bineph2fits", "File created by bineph2fits", &error) ;
  fits_write_key (FF, TDOUBLE, "TSTART", &ss1, "Start time of ephemeris", &error) ;
  fits_write_key (FF, TDOUBLE, "TSTOP", &ss2, "Stop time of ephemeris", &error) ;
  fits_write_key (FF, TDOUBLE, "TIMEDEL", &ss3, "Ephemeris interval", &error) ;
  fits_modify_key_lng (FF, "NAXIS2", nrow, "Number of rows", &error) ;
  fits_write_key (FF, TSTRING, "TIMESYS", "TT", "Time system of ephemeris", &error) ;
  fits_write_key (FF, TSTRING, "TIMEUNIT", "d", "Time is in days", &error) ;
  ss1 = 0.0 ;
  fits_write_key (FF, TDOUBLE, "JDREF", &ss1, "Time is in JD", &error) ;
  fits_write_chksum (FF, &error) ;

  fits_close_file (FF, &error) ;

  if ( error ) {
    fprintf (stderr, "bineph2fits: Failed writing third extension\n") ;
    fprintf (stderr, "bineph2fits: Error = %d\n", error) ;
  }

  fprintf (stderr, "bineph2fits: Wrote %d records to %s\n", nrow, ephemFile) ;

  exit (error) ;
}
