/* RCS: $Id: fBsub.c,v 3.1 1998/09/28 21:26:48 arots Exp arots $ */
/*-----------------------------------------------------------------------
 *
 *  fBsub.c
 *
 *  Date: 20 March 1997
 *
 *  Arnold Rots, USRA
 *
 *  fBsub allows subtraction of unpulsed background from a pha (fB) file.
 *  The average spectrum between phasein and phasemax is subtracted from
 *  all spectra in the Data column, and added to the Baseline column (so
 *  that fBsub can be run multiple times).  The errors are not touched.
 *
 *
 *----------------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fitsio.h>
#include <longnam.h>
#define NCHAN   4096

char *fB = "fBsub $Revision: 3.1 $" ;
int main (int argc, char *argv[])
{

  fitsfile *fits_file;
  char line[1024] ;
  char progName[256] ;
  char dataFile[1024] ;
  char outFile[1025] ;
  char keystrval[80] ;
  char comment[80] ;
  int error = 0 ;
  int htype ;
  long nrows, nch, i, j, n1, n2, n3, n4 ;
  double nulv = 0.0 ;
  int any ;
  double dnphase, x ;
  double ph1, ph2, ph3, ph4, phase ;
  double base[NCHAN] ;
  double spec[NCHAN] ;
  double nbas[NCHAN] ;
  double oerr[NCHAN] ;
  double nerr[NCHAN] ;

  int getopts (int, char **, char *, char *, char *,
	       double *, double *, double *, double *) ;

/*
 *  Decode command line options -----------------------------------------
 */
  printf ("# ") ;
  for (i=0; i<argc; i++)
    printf (" %s", argv[i]) ;
  printf ("\n") ;
  if ( getopts (argc, argv, progName, dataFile, outFile,
		&ph1, &ph2, &ph3, &ph4) )
    exit (1) ;
  while ( ph1 < 0.0 )
    ph1++ ;
  while ( ph1 >= 1.0 )
    ph1-- ;
  while ( ph2 > ph1 )
    ph2-- ;
  while ( ph2 <= ph1 )
    ph2++ ;
  if ( ph3 >= 0.0 ) {
    while ( ph3 >= 1.0 )
      ph3-- ;
    while ( ph4 > ph3 )
      ph4-- ;
    while ( ph4 <= ph3 )
      ph4++ ;
  }

/*
 *  Copy the input file -------------------------------------------------
 */
  if ( !strstr (outFile, ".") )
    strcat (outFile, ".pha") ;
  sprintf (line, "cp %s %s", dataFile, outFile) ;
  if ( system (line) ) {
    fprintf (stderr,"%s: Could not open input file %s or create output file %s\n",
	     progName, dataFile, outFile) ;
    exit (1) ;
  }

/*
 *  Fix up output file --------------------------------------------------
 */
  if ( fits_open_file (&fits_file, outFile, READWRITE, &error) ) {
    fprintf (stderr,"%s: Failed to re-open output file %s\n",
	     progName, outFile) ;
    exit (2) ;
  }
  fits_write_date (fits_file, &error) ;
  fits_write_chksum (fits_file, &error) ;

/*
 *  Check first extension -----------------------------------------------
 */
  fits_movabs_hdu (fits_file, 2, &htype, &error) ;
  fits_read_key (fits_file, TSTRING, "HDUCLAS5", keystrval, comment, &error) ;
  if ( strcmp (keystrval, "fB") ) {
    fprintf (stderr,"%s: Input file %s is of the wrong type\n",
	     progName, dataFile) ;
    fits_delete_file (fits_file, &error) ;
    exit (3) ;
  }
  fits_read_key_lng (fits_file,        "NAXIS2",    &nrows, comment, &error) ;
  fits_read_key_dbl (fits_file,      "PHASEDEL",  &dnphase, comment, &error) ;
  fits_read_key_lng (fits_file,      "DETCHANS",      &nch, comment, &error) ;
  fits_read_col (fits_file, TDOUBLE, 2, 1, 1, 1, &nulv, &phase, &any, &error) ;
  n1 = (ph1 - phase) / dnphase + 1.5 ;
  n2 = (ph2 - phase) / dnphase + 1.5 ;
  if ( ph3 >= 0.0 ) {
    n3 = (ph3 - phase) / dnphase + 1.5 ;
    n4 = (ph4 - phase) / dnphase + 1.5 ;
  }
  else
    n3 = n4 = 0 ;
  if ( n2 > nrows ) {
    fprintf (stderr,"%s: Phase %f not contained in Input file %s\n",
	     progName, ph2, dataFile) ;
    fits_delete_file (fits_file, &error) ;
    exit (4) ;
  }
  if ( n4 > nrows ) {
    fprintf (stderr,"%s: Phase2 %f not contained in Input file %s\n",
	     progName, ph4, dataFile) ;
    fits_delete_file (fits_file, &error) ;
    exit (5) ;
  }

/*
 *  Get baseline correction ---------------------------------------------
 */
  for (i=0; i<nch; i++)
    base[i] = 0.0 ;
  for (j=n1; j<=n2; j++) {
    fits_read_col (fits_file, TDOUBLE, 4, j, 1, nch, &nulv, spec, &any, &error) ;
    for (i=0; i<nch; i++)
      base[i] += spec[i] ;
  }
  x = 1.0 / (n2 - n1 + 1) ;
  if ( n3 ) {
    for (j=n3; j<=n4; j++) {
      fits_read_col (fits_file, TDOUBLE, 4, j, 1, nch, &nulv, spec, &any, &error) ;
      for (i=0; i<nch; i++)
	base[i] += spec[i] ;
    }
    x = 1.0 / (n2 - n1 + n4 - n3 + 2) ;
  }
  for (i=0; i<nch; i++)
    base[i] *= x ;
  
/*
 *  Set new baseline ----------------------------------------------------
 */
  fits_read_col (fits_file, TDOUBLE, 6, 1, 1, nch, &nulv, nbas, &any, &error) ;
  for (i=0; i<nch; i++) {
    nbas[i] += base[i] ;
    nerr[i] = nbas[i] * (1.0 + x) ;
  }

/*
 *  Enter corrections ---------------------------------------------------
 */
  for (j=1; j<=nrows; j++) {
    fits_read_col (fits_file, TDOUBLE, 4, j, 1, nch, &nulv, spec, &any, &error) ;
    for (i=0; i<nch; i++) {
      spec[i] -= base[i] ;
      oerr[i] = sqrt (spec[i] + nerr[i]) ;
    }
    fits_write_col_dbl (fits_file, 4, j, 1, nch, spec, &error) ;
    fits_write_col_dbl (fits_file, 5, j, 1, nch, oerr, &error) ;
    fits_write_col_dbl (fits_file, 6, j, 1, nch, nbas, &error) ;
  }
  
/*
 *  Last fixes and close ------------------------------------------------
 */
  fits_write_date (fits_file, &error) ;
  fits_modify_key_str (fits_file, "CREATOR", fB, "&", &error) ;
  fits_write_chksum (fits_file, &error) ;

  fits_close_file (fits_file, &error) ;

  exit (error) ;
}

/*-----------------------------------------------------------------------
 *
 *  getopts decodes command line options for faseBin.
 *
 *----------------------------------------------------------------------*/

int getopts (int argc, char **argv, char *progName, char *dataFile,
	     char *outFile, double *ph1, double *ph2, double *ph3, double *ph4)
{
/*
 *  Initialize ----------------------------------------------------------
 */
  dataFile[0] = 0 ;
  outFile[0] = 0 ;
  *ph2 = -100.0 ;
  *ph3 = -100.0 ;
  *ph4 = -100.0 ;
  strcpy (progName, *argv) ;
  argv++ ;
  argc-- ;

/*
 *  Decode --------------------------------------------------------------
 */
  while ( argc ) {
    if ( !strcmp (*argv, "-i") ) {
      strcpy (dataFile, *(++argv)) ;
      argc-- ;
    }
    else if ( !strcmp (*argv, "-o") ) {
      strcpy (outFile, *(++argv)) ;
      argc-- ;
    }
    else if ( !strcmp (*argv, "-phase2") ) {
      *ph3 = atof (*(++argv)) ;
      argc-- ;
      *ph4 = atof (*(++argv)) ;
      argc-- ;
    }
    else if ( !strcmp (*argv, "-phase") ) {
      *ph1 = atof (*(++argv)) ;
      argc-- ;
      *ph2 = atof (*(++argv)) ;
      argc-- ;
    }
    argv++ ;
    argc-- ;
  }

/*
 *  Check and return ----------------------------------------------------
 */
  if ( !dataFile[0] || !outFile[0] || ( *ph2 < 0.0 ) ) {
    fprintf (stderr, "fBsub - Usage:\n    %s ",  progName) ;
    fprintf (stderr, "-i inputFile -o outputFile -phase phmin phmax\n") ;
    fprintf (stderr, "                          [-phase2 phmin phmax]\n") ;
    return 1 ;
  }
  else
    return 0 ;
}
